"""
Key Profile Based Self-Organizing Maps

Self-organizing maps trained on pitch class usage profiles from literature.

A self-organizing map can be trained on key profile with 'major' and 'minor'
data fields. The caller can define the decay rate and neighborhood
function that the profile is trained on.
A projection of a pitch-class distribution (pc_projection) onto a
self-organizing map given the constituent input weights (input_weights) and
their corresponding pitch-class distribution weights (pcdist), is defined for
all valid row, column = i, j as
pc_projection[i, j, :] = sum(input_weights[i, j, k] * pcdist[k] for all k in range(12)).

When visualizing a projection of a pitch-class profile onto a trained
self-organizing map, there are 24 key labels scattered across the map.
The positions of these key labels (upper-case for major, lower-case for minor)
are determined by the position of the best matching unit of the corresponding
pitch profile in the trained map.

Reference
---------
Toiviainen, P. & Krumhansl, C. L. (2003). Measuring and modeling
real-time responses to music: the dynamics of tonality induction.
Perception, 32(6), 741-766.
"""

import math
import os

# for function types
from collections.abc import Callable
from typing import List, Optional, Tuple, Union

import matplotlib.colors as mcolors
import matplotlib.pyplot as plt
import numpy as np
from matplotlib.animation import FuncAnimation

# from matplotlib.axes import Axes
from matplotlib.figure import Figure

from amads.algorithms.norm import euclidean_distance
from amads.core.pitch import CHROMATIC_NAMES
from amads.pitch.key import profiles as prof


def zero_SOM_init(shape: Tuple[int, int, int]) -> np.array:
    """
    Constructs all-zero self-organizing map weights.
    """
    # zero SOM init...
    return np.zeros(shape)


def random_SOM_init(shape: Tuple[int, int, int]) -> np.array:
    """
    Constructs random self-organizing map weights.

    Initial values are between 0 and 0.5 inclusive.
    """
    return np.random.rand(*shape) / 2


def handcrafted_SOM_init(shape: Tuple[int, int, int]) -> np.array:
    """
    Bespoke initialization for the pretrained weights.

    These initial weights are intended to bias the resulting SOM
    to a nicely symmetrical map with well-placed keys.
    """
    if shape != (
        *KeyProfileSOM._default_output_dimensions,
        KeyProfileSOM._input_length,
    ):
        raise ValueError(f"invalid shape {shape} for handcrafted SOM")

    kkprof = prof.krumhansl_kessler
    list_of_majors = kkprof["major"].normalize().as_canonical_matrix()
    list_of_minors = kkprof["minor"].normalize().as_canonical_matrix()

    max_row, max_col, _ = shape

    init_weights = np.zeros(shape)

    # per "cell" arrangement, where each cell contains a major key label and its
    # corresponding minor key label horizontally adjacent to it

    # arithmetic for a 4x3 grid of cells in a 24x36 SOM
    row_multiplier = 6
    max_row_cells = max_row // row_multiplier
    row_offset = 3
    # twice to accomodate a cell containing 12
    col_multiplier = 6 * 2
    max_col_cells = max_col // col_multiplier
    col_major_offset = 3
    col_minor_offset = col_major_offset + 6

    key_idx = 0
    # the end result should be a rectangular grid in the labels
    for row_cell_idx in range(max_row_cells):
        for col_cell_idx in range(max_col_cells):
            # imprint major key
            major_col_idx = col_cell_idx * col_multiplier + col_major_offset
            major_row_idx = row_cell_idx * row_multiplier + row_offset
            init_weights[major_row_idx, major_col_idx] = list_of_majors[key_idx]

            # imprint minor key
            minor_col_idx = col_cell_idx * col_multiplier + col_minor_offset
            minor_row_idx = row_cell_idx * row_multiplier + row_offset
            init_weights[minor_row_idx, minor_col_idx] = list_of_minors[key_idx]

            # increment key_idx
            key_idx += 1

    assert key_idx == 12

    return init_weights


def keysom_inverse_decay(idx: int) -> float:
    """
    Computes inverse decay global learning rate for a given iteration.

    Inverse to the current learning iteration...
    """
    assert idx >= 0
    if idx == 0:
        return 1
    else:
        return 1 / (2 * idx)


def keysom_stepped_inverse_decay(idx: int) -> float:
    """
    Compute stepped decay global learning rate for a given iteration.

    Inverse to a stepped multiplier of the current learning iteration...

    In this case it's stepped to allow all inputs to deterministically
    pass during training (for 12 major and 12 minor key profile entries
    specifically)
    """
    step = idx // (12 * 2)
    if step == 0:
        return 1
    else:
        return 1 / (2 * step)


def keysom_stepped_log_inverse_decay(idx: int) -> float:
    """
    Compute stepped log inverse decay global learning rate for a given iteration.

    Log inverse to a stepped multiplier of the current learning iteration...

    In this case it's stepped to allow all inputs to deterministically
    pass during training (for 12 major and 12 minor key profile entries
    specifically)

    So why does inverse log work well? I like to think of it as the summation
    of traversing all nodes of a subtree of an imaginary sparse information tree.
    Where each layer of the information tree provides inverse decaying
    returns to the whole tree.
    This justification is very stretched though.
    Namely, each additional data point fed provides an opportunity to add
    another "symbol" to the intrinsic learned "alphabet" of the internal
    representation of the SOM.
    """
    step = idx // (12 * 2)
    if step == 0:
        return 1
    else:
        return 1 / math.log2(step + 2)


def keysom_centroid_euclidean(
    coord: Tuple[int], best_match: Tuple[int], shape: Tuple[int], idx: int
) -> float:
    """
    Neighborhood propagation update function.

    Exponential decay based off of Eucliean distance on a 2-D plane
    assuming the SOM output layer is a 2-D flat plane
    """
    distance = euclidean_distance(coord, best_match, False)

    # 0.95 is purely empirical. Honestly another value might be better
    if distance == 0:
        return 1
    else:
        return (0.95) ** distance


def keysom_toroid_euclidean(
    coord: Tuple[int], best_match: Tuple[int], shape: Tuple[int], idx: int
) -> float:
    """
    Neighborhood propagation update function.

    Exponential decay based off of Eucliean distance on a toroid
    assuming the SOM output layer is a projection of a toroid onto a
    2-D flat plane
    """
    num_rows, num_cols, input_length = shape
    # these are easy to follow since i, j are convention
    # and so is i0, j0
    i, j = coord
    i0, j0 = best_match
    diff_row = abs(i - i0)
    diff_col = abs(j - j0)
    toroid_diff_row = min(diff_row, num_rows - diff_row)
    toroid_diff_col = min(diff_col, num_cols - diff_col)
    distance = math.sqrt(toroid_diff_row**2 + toroid_diff_col**2)
    if distance == 0:
        return 1
    else:
        return (0.9) ** distance


def keysom_toroid_clamped(
    coord: Tuple[int], best_match: Tuple[int], shape: Tuple[int], idx: int
) -> float:
    """
    Neighborhood propagation update function.

    Same behavior as keysom_toroid_euclidean (see for more details),
    except distances past a certain radius is clamped to 0.0001.
    """
    num_rows, num_cols, input_length = shape
    # these are easy to follow since i, j are convention
    # and so is i0, j0
    i, j = coord
    i0, j0 = best_match
    diff_row = abs(i - i0)
    diff_col = abs(j - j0)
    toroid_diff_row = min(diff_row, num_rows - diff_row)
    toroid_diff_col = min(diff_col, num_cols - diff_col)
    distance = math.sqrt(toroid_diff_row**2 + toroid_diff_col**2)

    # same logic applies to clamp radius as the global learning decay rate.
    # Namely, each additional data point fed provides an opportunity to add
    # another "symbol" to the intrinsic learned "alphabet" of the internal
    # representation of the SOM.
    # diminishing returns...
    radius = 36.0 * (1 / math.log2(idx // 24 + 2))

    if distance > radius:
        return 0.0001
    elif distance == 0:
        return 1.0
    else:
        return (0.9) ** distance


class KeyProfileSOM:
    """
    The primary use-case for Key Profile SOM is to extract tonal features of a
    pair of major and minor pitch profiles and project them onto a 2-D map.

    We provide methods to allows users to do the following:

    1. For initialization, supply the structure of the key profile SOM.

    2. For training, the following is customizeable:

        2.1 Customize the key profile used for training data. Only
    the 'major' and 'minor' PitchProfile attributes within the supplied key
    profile are used.
    If either attribute does not exist, a value error is raised before training.

        2.2 Customize training behavior by supplying functions denoting both the
    neighborhood propagation function and global decay function.

        2.3 Record a log of information to track the training process. This
    feature is not implemented completely.

    3. Projection of a supplied pitch-class distribution onto a trained SOM.

    4. Visualization facilities, primarily:

        4.1 Visualization of projection from a single pitch-class distribution
    onto the internal SOM.

        4.2 Animation of projections from a sequence of pitch-class distributions
    onto the internal SOM.

    The internal data unique to each KeyProfileSOM object is as follows:

    1. SOM structural specifications (output layer dimensions), stored as a
    2-tuple of ints specifying the width and height of the SOM output layer.

    2. The trained SOM according to the structural specifications,
    stored as a 3-D numpy array with shape
    (output_width, output_height, input_length).

    3. A list of 2-D coordinates for the trained SOM, where each list element at
    index i specifies the location of the ith label in the trained SOM.
    The coordinate is the BMU of the ith pitch's corresponding pitch-class
    profile in the trained SOM.
    Note that capital pitches denote major key pitches, while lower-case pitches
    denote minor key pitches.

    4. Training log containing simple training information per update instance
    of the SOM (for debugging purposes if logging is turned on during training).

    5. A name (string identifier) for the SOM that can be optionally specified.

    Attributes
    ----------
    _input_length : int, class attribute
        the number of pitches in a pitch-class distribution, or the number
        of tonal pitches in a western music system
    _default_output_dimensions : Tuple[int], class attribute
        output dimensions for the pretrained SOM in the original MATLAB version
        of miditoolbox
    _labels : List[str], class attribute
        list of strings denoting the pitch labels

    Examples
    --------
    Load a set of pretrained weights and visualize the pitch-class distribution
    of a small score:
    >>> from amads.core.basics import Score
    >>> score = Score.from_melody([60, 62, 64, 65, 67, 69, 71, 72])
    >>> from amads.pitch.pcdist1 import pitch_class_distribution_1
    >>> pcdist_of_score = pitch_class_distribution_1(score)
    >>> example_SOM = pretrained_weights_script()
    >>> _ = example_SOM.project_and_visualize(tuple(pcdist_of_score.data),
    ...                                       show=False)

    Train a set of weights from a key profile with 'major' and 'minor'
    attributes with supplied training parameters:
    >>> training_profile = prof.krumhansl_kessler # from key/profiles.py
    >>> test_SOM = KeyProfileSOM() # default output dimensions used
    >>> _ = test_SOM.train_SOM(training_profile) # use default parameters

    """

    # corresponds to the number of weights in a pitch-class distribution
    _input_length = 12
    _default_output_dimensions = (24, 36)
    # labels (major and minor)
    _labels = CHROMATIC_NAMES + [pitch.lower() for pitch in CHROMATIC_NAMES]

    def __init__(
        self, output_layer_dimensions: Tuple[int] = _default_output_dimensions
    ):
        self.SOM_output_dims = output_layer_dimensions
        self.SOM = None
        # best matching units to each of the corresponding coordinates
        self.label_coord_list = None
        self.log_info = []
        self.name = None

    @classmethod
    def save_trained_SOM(
        cls,
        obj: "KeyProfileSOM",
        dir_path: str = "./",
        file_name: Optional[str] = None,
    ):
        """
        Save a trained key profile SOM.

        Raises a value exception if the object
        does not contain a proper trained SOM or the directory path is not valid.

        Parameters
        ----------
        obj: KeyProfileSOM
            Key Profile SOM object containing a trained SOM
        dir_path: str
            Path to directory to store the trained SOM (in npz format)
        file_name: Optional[str]
            Optional file name argument to save the trained SOM
        """

        if file_name is None:
            file_name = f"{obj.name}_data.npz"

        file_path = os.path.join(dir_path, file_name)

        if obj.SOM is None or not obj.label_coord_list:
            raise ValueError("input SOM is not trained!")

        np.savez(
            file_path,
            SOM=obj.SOM,
            name=np.array(obj.name),
            label_coords=np.array(obj.label_coord_list),
        )

    @classmethod
    def from_trained_SOM(
        cls, file_path: str = "./amads/pitch/key/KrumhanslKessler_SOM_data.npz"
    ) -> "KeyProfileSOM":
        """
        Create a new KeyProfileSOM object containing the trained KeyProfileSOM.

        Data is loaded from the specified file.

        Parameters
        ----------
        file_path: str
            Path to directory containing stored SOM (in npz format)

        Returns
        -------
        KeyProfileSOM
            Key Profile SOM object containing the trained SOM from the file
        """
        load_table = np.load(file_path)
        SOM = load_table["SOM"]
        (dim0, dim1, _) = SOM.shape
        output_dims = (dim0, dim1)
        name = str(load_table["name"])
        label_coord_list = [
            tuple(coord) for coord in load_table["label_coords"]
        ]

        obj = KeyProfileSOM(output_dims)
        obj.SOM = SOM
        obj.name = name
        obj.label_coord_list = label_coord_list
        obj.vmin = np.min(obj.SOM)
        obj.vmax = np.max(obj.SOM)

        return obj

    def update_SOM(
        self,
        best_match: Tuple[int],
        input_data: np.array,
        idx: int,
        neighborhood: Callable[
            [Tuple[int], Tuple[int], Tuple[int], int], float
        ],
        global_decay: Callable[[int], float],
    ) -> "KeyProfileSOM":
        """
        Update the SOM on the input data based off of the best matching unit.

        Uses the current global training iteration.

        Parameters
        ----------
        best_match: Tuple[int]
            Coordinate of the best-matching node in the output layer of the
            self-organizing map and its corresponding connector weights
            (to the input)
        input_data: np.array
            data vector that was selected to train on for the current
            training iteration
        idx: int
            Current training iteration in training session.
        neighborhood: Callable[[Tuple[int], Tuple[int], Tuple[int], int], float]
            Neighborhood function, denoting the update rate component depending
            on coordinate differences to the best matching unit and training
            iteration.
        global_decay: Callable[[int], float]
            Global decay function, denoting the update rate component dependent
            solely on training iteration.

        Returns
        -------
        KeyProfileSOM
            Current object
        """
        if input_data.shape != (KeyProfileSOM._input_length,):
            raise ValueError(
                f"input {input_data} is of invalid shape {input_data.shape}"
            )

        dim0, dim1, input_length = self.SOM.shape
        if input_length != KeyProfileSOM._input_length:
            raise ValueError(
                f"Corrupted SOM =\n{self.SOM}\nof shape {self.SOM.shape}"
            )
        if len(best_match) != 2:
            raise ValueError(f"Invalid best matching unit coords {best_match}")

        # update all weights in the SOM based on competitive learning
        # (where the only things that truly matter in a training iteration
        # is the BMU and the input data)
        for i in range(dim0):
            for j in range(dim1):
                rate = global_decay(idx) * neighborhood(
                    (i, j), best_match, self.SOM.shape, idx
                )
                self.SOM[i, j, :] = (1 - rate) * self.SOM[
                    i, j, :
                ] + rate * input_data

        return self

    def find_best_matching_unit(self, input_data: np.array) -> Tuple[int]:
        """
        Find best matching unit given a self-organizing map and input data.

        Finds the coordinate of the output node whose weights has the smallest
        Euclidean distance from the input data.

        Parameters
        ----------
        input_data: np.array
            1-D data vector of input length containing the input weights

        Returns
        -------
        Tuple[int]
            Coordinates of the node that has the connector weights with the
            smallest Euclidean distance to the input data
        """
        # input data length needs to match
        if input_data.shape != (KeyProfileSOM._input_length,):
            raise ValueError(
                f"input {input_data} is of invalid shape {input_data.shape}"
            )

        dim0, dim1, input_length = self.SOM.shape
        if input_length != KeyProfileSOM._input_length:
            raise ValueError(
                f"Corrupted SOM =\n{self.SOM}\nof shape {self.SOM.shape}"
            )

        best_i, best_j = 0, 0
        best_distance = euclidean_distance(
            input_data, self.SOM[best_i, best_j, :], False
        )
        for i in range(dim0):
            for j in range(dim1):
                distance = euclidean_distance(
                    input_data, self.SOM[i, j, :], False
                )
                if best_distance > distance:
                    best_i, best_j = i, j
                    best_distance = distance
        return (best_i, best_j)

    def _data_selector(
        self, training_data: np.array, training_idx: int
    ) -> Tuple[int, np.array]:
        """
        Internal data selector amongst training inputs represented as a list of
        canonical matrices (See the PitchProfile class in key/profiles.py for
        more detials on canonical matrices).

        This training data selector very specific to amads implementation
        of PitchProfile (See key/profiles.py for more details on the
        PitchProfile class).

        Parameters
        ----------
        Training data:
            2-D numpy array where each row is a training data input, and each column
            index correspond to the indices in a chromatic scale
        training_idx:
            Index of current training iteration

        Returns
        -------
        np.array[float]
            An input vector (with the weights of a normalized pitch profile)
            that is a 1-D numpy vector.
        """

        num_data, input_length = training_data.shape
        if input_length != KeyProfileSOM._input_length:
            raise ValueError(
                f"invalid training data dimensions {training_data.shape},"
                " expected (num_data, 12))"
            )

        # instead of random.randrange(0, num_data), let's just see the
        # deterministic version instead...

        # first scan through all major keys in the chromatic order of circle of
        # fifths
        # then scan through all minor keys in the same chromatic order
        if training_idx % 24 < 12:
            access_idx = (training_idx * 5) % 12
            return access_idx, training_data[access_idx, :]
        else:
            access_idx = 12 + ((training_idx - 12) * 5) % 12
            return access_idx, training_data[access_idx, :]

    def _log_training_iteration(
        self,
        training_idx: int,
        data_idx: int,
        bmu: Tuple[int],
    ):
        """
        logs a training iteration. This function can be used to visualize
        a training iteration.

        Parameters
        ----------
        training_idx: int
            current training iteration

        data_idx: int
            index of the selected data row in the data matrix

        bmu: Tuple[int]
            coordinate of the best matching unit for the selected data during
            the current training iteration.
        """
        self.log_info.append((training_idx, data_idx, bmu))
        return

    def train_SOM(
        self,
        profile: prof.KeyProfile = prof.krumhansl_kessler,
        max_iterations: int = 24 * 64,
        neighborhood: Callable[
            [Tuple[int], Tuple[int], Tuple[int], int], float
        ] = keysom_toroid_clamped,
        global_decay: Callable[[int], float] = keysom_stepped_log_inverse_decay,
        weights_initialization: Callable[
            [Tuple[int, int, int]], np.array
        ] = random_SOM_init,
        log_training: bool = False,
    ) -> "KeyProfileSOM":
        """
        Train a self-organizing map using the given training data and parameters.

        Parameters
        ----------
        profile: prof.KeyProfile
            The key profile to use for analysis.
        max_iterations: int
            The number of iterations to train the self-organizing map for

            Neighborhood function, denoting the update rate component dependent
            on coordinate differences to the best matching unit and training
            iteration.
        global_decay: Callable[[int], float]
            Global decay function, denoting the update rate component dependent
            solely on training iteration.
        weights_initialization: Callable[[Tuple[int, int, int]], np.array]
            SOM weights initialization function, returning a numpy array of the
            initial SOM weights dependent on its input shape.
        log_training: bool
            Indicator flag for whether or not to keep a semi-detailed log of the
            training process

        Returns
        -------
        KeyProfileSOM
            Current object
        """
        attribute_names = ["major", "minor"]

        data_multiplier = 6

        # multiplied by 6 which is the expected value of randomly initializing
        # the values of each neuron's map weights to a random value between
        # [0, 1]
        # think of this as "normalizing" the input data to the same scale as
        # the original map weights

        # this will not affect the visualization of projections of a pitch-class
        # distribution in any way, since we can simply multiply the SOM
        # globally by the requisite multiplier after it is trained
        # in order to obtain normalized neuron weights.
        list_of_canonicals = [
            profile[attribute].normalize().as_canonical_matrix()
            * data_multiplier
            for attribute in attribute_names
        ]
        self.name = profile.name + "_SOM"

        # stack into matrix representation to satisfy _data_selector argument
        # specification
        # Additionally, training data here is ordered (per row) in chromatic
        # scale order, first in major pitch profile weights then minor pitch
        # profile weights.
        training_data = np.vstack(list_of_canonicals)

        # 12 is the input length
        # 36 is data width/feature width/something else?
        # 24 is the 12 major and 12 minor keys for the profile data
        #
        # SOM indices have been rearranged from matlab version for convenience
        # in this implementation
        self.SOM = weights_initialization(
            (*self.SOM_output_dims, KeyProfileSOM._input_length)
        )

        for training_idx in range(max_iterations):
            data_idx, data_vector = self._data_selector(
                training_data, training_idx
            )
            best_match = self.find_best_matching_unit(data_vector)
            self.update_SOM(
                best_match=best_match,
                input_data=data_vector,
                idx=training_idx,
                neighborhood=neighborhood,
                global_decay=global_decay,
            )
            if log_training:
                self._log_training_iteration(training_idx, data_idx, best_match)

        self.label_coord_list = []
        # need to find BMU to each of the key profile input vectors in the trained SOM
        # since training_data is already ordered properly, we just need to find BMU
        # over all the inputs
        for label, input in zip(KeyProfileSOM._labels, training_data):
            best_match = self.find_best_matching_unit(input)
            self.label_coord_list.append(best_match)
            # print(f"label: {label}, best_match: {best_match}")

        # setting colorbar scale here
        self.vmin = np.min(self.SOM)
        self.vmax = np.max(self.SOM)

        return self

    # TODO: should I have an additional function here to set visualization
    # state so that it's persistent across multiple plots?

    def project_input_onto_SOM(self, input_data: np.array) -> np.array:
        """
        Compute the resulting projection weights of the input on a trained SOM.

        Parameters
        ----------
        input_data: np.array
            1-D data vector of input length containing the input weights

        Returns
        -------
        np.array[float]
            Returns a 2-D numpy array that contains the projection of the input
            data onto the self-organizing map.
        """
        # input data length needs to match
        if input_data.shape != (KeyProfileSOM._input_length,):
            raise ValueError(
                f"input {input_data} is of invalid shape {input_data.shape}"
            )

        dim0, dim1, input_length = self.SOM.shape
        if input_length != KeyProfileSOM._input_length:
            raise ValueError(
                f"Corrupted SOM =\n{self.SOM}\nof shape {self.SOM.shape}"
            )
        # projection of input data onto current self-organizing map
        # matrix multiplication (tensor extension)
        application = self.SOM @ input_data
        assert application.shape == (dim0, dim1)
        return application

    def project_and_visualize(
        self,
        input: Tuple[float],
        has_legend: bool = True,
        scaled_legend: bool = False,
        font_size: Optional[Union[float, str]] = None,
        color_map: Optional[Union[str, mcolors.LinearSegmentedColormap]] = None,
        show: bool = True,
    ) -> Tuple[np.array, Figure]:
        """
        Project a pitch-class distribution and visualizes it.

        Parameters
        ----------
        input: Tuple[float]
            a singular pitch-class distribution, in which case the visualization
            is simply a heatmap of its projection onto the trained SOM.
        has_legend: bool
            Whether or not the plot should include a color legend
        scaled_legend: bool
            Whether or not the color legend scales with the projection's minimum
            and maximum, or (by default) scales with the trained SOM's global
            minimum and maximum.
        font_size: Optional[Union[float, str]] = None,
            Font size, either:
            (1) Font size of the labels (in points) or a string option from
            matplotlib
            (2) None for the default font size provided by matplotlib
        color_map: Optional[Union[str, mcolors.LinearSegmentedColormap]]
            Color map, either:
             (1) a color map provided by the matplotlib package
             (2) a custom linear segmented colormap
             (3) None for the default color scheme provided by matplotlib
        show: bool
            Whether or not we suspend execution and display the plot before
            returning from this function

        Returns
        -------
        Tuple[np.array, Figure]
            Returns a tuple consisting of:

            1. the 2-D numpy array that contains the projection of the input
                data onto the self-organizing map.
            2. Matplotlib figure that contains the axes with a plot of the
                projection
        """
        if not input:
            raise ValueError("empty input not allowed")
        if isinstance(input[0], Tuple):
            raise ValueError("only takes pitch-class distribution data")
        # prep data
        projection = self.project_input_onto_SOM(np.array(input))

        dim0, dim1, _ = self.SOM.shape
        assert projection.shape == (dim0, dim1)

        fig, ax = plt.subplots()

        # there should be some thought put into the actual interpolation formula
        # cax = ax.contourf(projection)
        cax = ax.imshow(
            projection,
            aspect="auto",
            origin="lower",
            interpolation="nearest",
            cmap=color_map,
        )
        assert len(self.label_coord_list) == len(KeyProfileSOM._labels)

        # key labels in the plot
        for (i, j), label in zip(self.label_coord_list, KeyProfileSOM._labels):
            ax.text(
                j,
                i,
                label,
                ha="center",
                va="center",
                color="w",
                fontsize=font_size,
            )

        # legend
        if has_legend:
            actual_vmin, actual_vmax = self.vmin, self.vmax

            if scaled_legend:
                actual_vmin = np.min(projection)
                actual_vmax = np.max(projection)

            cax.set_clim(actual_vmin, actual_vmax)
            fig.colorbar(cax, ax=ax, label="Proportion")

        if show:
            plt.show()

        return projection, fig

    def project_and_animate(
        self,
        input_list: List[Tuple[float]],
        has_legend: bool = True,
        font_size: Union[float, str] = 10.0,
        color_map: Optional[Union[str, mcolors.LinearSegmentedColormap]] = None,
        show: bool = True,
    ) -> Tuple[List[np.array], FuncAnimation]:
        """
        Animate a collection of pitch-class distributions.

        Parameters
        ----------
        input_list: List[Tuple[float]]
            a list of pitch-class distributions, in which case the visualization
            is an animation of the sequence of projections of the pitch-class
            distributions onto the trained SOM.
        has_legend: bool
            Whether or not the plot should include a color legend
        font_size: Union[float, str]
            Font size, either:

            1. Font size of the labels (in points) or a string option from
                matplotlib
            2. None for the default font size provided by matplotlib
        color_map: Optional[Union[str, mcolors.LinearSegmentedColormap]]
            Color map, either:

             1. a color map provided by the matplotlib package
             2. a custom linear segmented colormap
                 (see matplotlib.LinearSegmentedColormap for more details)
             3. None for the default color scheme provided by matplotlib
        show: bool
            Whether or not we suspend execution and display the plot before
            returning from this function

        Returns
        -------
        Tuple[List[np.array], ArtistAnimation]

            1. A list of 2-D numpy arrays that contain the sequence of
                projections from the list of input data onto the
                self-organizing map
            2. The artist animation object of these data
        """
        # visualize
        projection_list = [
            self.project_input_onto_SOM(np.array(input)) for input in input_list
        ]
        print(projection_list)
        if not projection_list:
            print("Warning! No distributions provided to animate!")
            return

        fig, ax = plt.subplots()

        cax = ax.imshow(
            projection_list[0],
            aspect="auto",
            origin="lower",
            interpolation="nearest",
            cmap=color_map,
        )
        cax.set_clim(self.vmin, self.vmax)
        # key labels in the plot
        for (i, j), label in zip(self.label_coord_list, KeyProfileSOM._labels):
            ax.text(
                j,
                i,
                label,
                ha="center",
                va="center",
                color="w",
                fontsize=font_size,
            )
        if has_legend:
            fig.colorbar(cax, ax=ax, label="Score Expectation", ticks=None)

        def frame_func(frame_idx):
            idx = frame_idx % len(projection_list)
            cax.set_data(projection_list[idx])

        ani = FuncAnimation(
            fig,
            frame_func,
            frames=len(input_list),
            interval=500,
            repeat=True,
            repeat_delay=2000,
        )

        if show:
            plt.show()

        return projection_list, ani


def pretrained_weights_script() -> KeyProfileSOM:
    """
    Simple script that generates a SOM from a hand-crafted initial SOM.

    This gives us a SOM with key labels in a determinstic grid adjacent to the
    axes of the grid.

    Returns
    -------
    KeyProfileSOM
        Object with training weights
    """

    # for the pretrained version, hand-craft initialization value for the
    # SOM so that we get the rotation and orientation desired for the resulting
    # trained weights
    obj = KeyProfileSOM()
    if obj.SOM_output_dims != KeyProfileSOM._default_output_dimensions:
        raise RuntimeError("invalid output dimensions for default SOM")

    obj.train_SOM(weights_initialization=handcrafted_SOM_init)

    return obj



"""
This module provides functionality for measuring the complexity of discrete sequences
using the LZ77 compression algorithm.

<small>**Author**: Huw Cheston (2025)</small>

References
----------

  - Ziv, J., & Lempel, A. (1977). A universal algorithm for sequential data compression.
    *IEEE Transactions on Information Theory*, 23(3), 337–343.
    https://doi.org/10.1109/TIT.1977.1055714

  - Cheston, H., Schlichting, J. L., Cross, I., & Harrison, P. M. C. (2024).
    Rhythmic qualities of jazz improvisation predict performer identity and style
    in source-separated audio recordings. *Royal Society Open Science*, 11(11).
    https://doi.org/10.1098/rsos.240920

"""

from typing import Hashable, Iterable, Union

__author__ = "Huw Cheston"


def lz77_encode(input_list: list[Hashable]) -> list:
    """Runs the LZ77 compression algorithm over the input `data`,
    generating tuples of (distance, length, symbol)

    Parameters
    ----------
    input_list : list[Hashable]
        The input sequence to be compressed.

    Returns
    -------
    list
        A list of tuples of (distance, length, symbol)
        representing the LZ77 encoded data.
    """

    # Catch sequences that have a length of zero
    if len(input_list) == 0:
        raise ValueError("Cannot encode an empty sequence!")
    encoded = []
    i = 0
    while i < len(input_list):
        best_length = 0
        best_distance = 0
        # Search the entire processed portion (positions 0 to i-1)
        for j in range(0, i):
            length = 0
            # Allow overlapping matches: compare while within input bounds.
            while (
                (i + length < len(input_list))
                and (j + length < len(input_list))
                and (input_list[j + length] == input_list[i + length])
            ):
                length += 1
            if length > best_length:
                best_length = length
                best_distance = i - j
        # If a match was found, output a match token.
        if best_length > 0:
            encoded.append((best_distance, best_length, input_list[i]))
            i += best_length
        else:
            # No match: output a literal token with length 1.
            encoded.append((0, 1, input_list[i]))
            i += 1
    return encoded


def lz77_decode(encoded: list[tuple[int, int, Hashable]]) -> list[Hashable]:
    """Decode a list of LZ77 tokens, each of which are 3-tuples with form (distance, length, symbol)

    Parameters
    ----------
    encoded : list[tuple[int, int, Hashable]]
        A list of LZ77 encoded tokens.

    Returns
    -------
    list[Hashable]
        The decoded sequence.
    """

    # Catch sequences that have a length of zero
    if len(encoded) == 0:
        raise ValueError("Cannot decode an empty sequence!")
    decoded = []
    for token in encoded:
        distance, length, symbol = token
        # Literal token: simply append the symbol.
        if distance == 0:
            decoded.append(symbol)
        # Match token: copy 'length' characters from the already decoded list,
        # starting at position len(decoded) - distance.
        else:
            start = len(decoded) - distance
            for i in range(length):
                decoded.append(decoded[start + i])
    return decoded


def lz77_complexity(
    sequence: Iterable[Hashable], normalized: bool = False
) -> Union[int, float]:
    """
    Compresses a discrete sequence using the LZ77 algorithm and returns the length of the compressed output.

    This function applies LZ77 compression to a sequence of hashable elements (e.g., strings, floats, integers).
    Higher compression lengths indicate greater complexity in the sequence.

    Parameters
    ----------
    sequence : Iterable[Hashable]
        A discrete sequence, such as pitch classes or inter-onset intervals.

    normalized : Optional(bool)
        If True, the result is expressed relative to the input size, where 1.0 indicates
        maximum complexity (i.e., no compression is possible). Defaults to False.

    Returns
    -------
    int | float:
        The length of the compressed sequence, either as a raw integer or a normalized float.
    """

    # Compress the sequence
    sequence = list(sequence)
    compressed = lz77_encode(sequence)
    # Express with relation to length of input string if required: returns a float
    if normalized:
        return len(compressed) / len(sequence)
    # Otherwise, take the length of the compressed string as the complexity of the input: returns an int
    else:
        return len(compressed)



"""Calculates the Interpolation Contour of a melody, along with related features, as
implemented in the FANTASTIC toolbox of Müllensiefen (2009) [1]
(features 23–27).
"""

__author__ = "David Whyatt"

from typing import Optional, Sequence

import numpy as np

from amads.core.basics import Score


class InterpolationContour:
    """Class for calculating and analyzing interpolated contours of melodies.

    As implemented in the FANTASTIC toolbox of Müllensiefen (2009) [1] (as
    features 23–27). This representation was first formalised by Steinbeck
    (1982) [2], and informed a variant of the present implementation in
    Müllensiefen & Frieler (2004) [3].

    Includes a modified version of the FANTASTIC method that is better
    suited to short melodies than the original implementation. This
    'AMADS' method defines turning points using reversals, and is the
    default method. All features are returned for either method.

    An interpolation contour is produced by first identifying turning points
    in the melody, and then interpolating a linear gradient between each
    turning point. The resulting list of values represents the gradient
    of the melody at evenly spaced points in time.

    <small>**Author**: David Whyatt</small>
    """

    def __init__(
        self,
        score: Optional[Score] = None,
        onsets: Optional[Sequence[float]] = None,
        pitches: Optional[Sequence[int]] = None,
        method: str = "amads",
    ):
        """Initialize with pitch and time values.

        Parameters
        ----------
        score : Score, optional, default=None
            If `pitches` and `onsets` are provided, use them. If not and a `score` is use that.
        pitches : list[int]
            Pitch values in any numeric format (e.g., MIDI numbers).
        onsets : list[float]
            Onset onsets in any consistent, proportional scheme (e.g., seconds,
            quarter notes, etc.)
        method : str, optional
            Method to use for contour calculation, either "fantastic" or "amads".
            Defaults to "amads".
            The FANTASTIC method is the original implementation, and identifies
            turning points using contour extrema via a series of rules. The
            AMADS method instead identifies reversals for all melody lengths,
            and is the default method.

        Raises
        ------
        ValueError
            If neither `onsets` and `pitches` or a score parameter are provided.
            If the `onsets` and `pitches` parameters are not the same length.
            If method is not "fantastic" or "amads"

        Examples
        --------
        >>> happy_birthday_pitches = [
        ...     60, 60, 62, 60, 65, 64, 60, 60, 62, 60, 67, 65,
        ...     60, 60, 72, 69, 65, 64, 62, 70, 69, 65, 67, 65
        ... ]
        >>> happy_birthday_onsets = [
        ...     0, 0.75, 1, 2, 3, 4, 6, 6.75, 7, 8, 9, 10,
        ...     12, 12.75, 13, 14, 15, 16, 17, 18, 18.75, 19, 20, 21
        ... ]
        >>> ic = InterpolationContour(
        ...     pitches=happy_birthday_pitches,
        ...     onsets=happy_birthday_onsets,
        ...     method="fantastic",
        ... )
        >>> ic.direction_changes
        0.6
        >>> ic.class_label
        'ccbc'
        >>> round(ic.mean_gradient, 6)
        2.702857
        >>> round(ic.gradient_std, 6)
        5.65564
        >>> ic.global_direction
        1

        References
        ----------
         1. Müllensiefen, D. (2009). Fantastic: Feature ANalysis Technology
            Accessing STatistics (In a Corpus): Technical Report v1.5

         2. W. Steinbeck, Struktur und Ähnlichkeit: Methoden automatisierter
            Melodieanalyse. Bärenreiter, 1982.

         3. Müllensiefen, D. & Frieler, K. (2004). Cognitive Adequacy in the
            Measurement of Melodic Similarity: Algorithmic vs. Human Judgments
        """
        none_checks = (onsets is not None, pitches is not None)
        if any(none_checks) and not all(none_checks):
            raise ValueError(
                "onsets and pitches must be provided together, not one without the other."
            )

        if all(none_checks):
            if len(onsets) != len(pitches):
                raise ValueError(
                    f"onsets and pitches must have the same length, "
                    f"got {len(onsets)} and {len(pitches)}."
                )
            self.onsets = list(onsets)
            self.pitches = list(pitches)
        else:
            if score is None:
                raise ValueError(
                    "Provide either a Score or both onsets and pitches."
                )
            if not isinstance(score, Score):
                raise TypeError("Score should be a Score object.")

            self.onsets, self.pitches = self.get_onsets_and_pitches(score)

        if method not in ["fantastic", "amads"]:
            raise ValueError(
                f"Method must be either 'fantastic' or 'amads', got {method}"
            )
        self.method = method

        self.contour = self.calculate_interpolation_contour(
            pitches, onsets, method
        )

    def get_onsets_and_pitches(
        self, score: Score
    ) -> tuple[list[float], list[int]]:
        """Extract onset times and pitches from a Score object.

        Parameters
        ----------
        score : Score
            The Score object to extract data from

        Returns
        -------
        tuple[list[float], list[int]]
            A tuple containing (onset_times, pitch_values)
        """
        notes = score.get_sorted_notes()
        return [note.onset for note in notes], [note.key_num for note in notes]

    @staticmethod
    def _is_turning_point_fantastic(pitches: list[int], i: int) -> bool:
        """Helper method to determine if a point is a turning point in FANTASTIC method."""
        return any(
            [
                (pitches[i - 1] < pitches[i] and pitches[i] > pitches[i + 1]),
                (pitches[i - 1] > pitches[i] and pitches[i] < pitches[i + 1]),
                (
                    pitches[i - 1] == pitches[i]
                    and pitches[i - 2] < pitches[i]
                    and pitches[i] > pitches[i + 1]
                ),
                (
                    pitches[i - 1] < pitches[i]
                    and pitches[i] == pitches[i + 1]
                    and pitches[i + 2] > pitches[i]
                ),
                (
                    pitches[i - 1] == pitches[i]
                    and pitches[i - 2] > pitches[i]
                    and pitches[i] < pitches[i + 1]
                ),
                (
                    pitches[i - 1] > pitches[i]
                    and pitches[i] == pitches[i + 1]
                    and pitches[i + 2] < pitches[i]
                ),
            ]
        )

    @staticmethod
    def calculate_interpolation_contour(
        pitches: list[int], onsets: list[float], method: str = "amads"
    ) -> list[float]:
        """Calculate the interpolation contour representation of a melody [1].

        Returns
        -------
        list[float]
            Array containing the interpolation contour representation
        """
        if method == "fantastic":
            return InterpolationContour._calculate_fantastic_contour(
                pitches, onsets
            )

        return InterpolationContour._calculate_amads_contour(pitches, onsets)

    @staticmethod
    def _calculate_fantastic_contour(
        pitches: list[int], onsets: list[float]
    ) -> list[float]:
        """
        Calculate the interpolation contour using the FANTASTIC method.

        Utilises the helper function _is_turning_point_fantastic to identify
        turning points.
        """
        # Find candidate points
        candidate_points_pitch = [pitches[0]]  # Start with first pitch
        candidate_points_time = [onsets[0]]  # Start with first time

        # Special case for very short melodies
        if len(pitches) in [3, 4]:
            for i in range(1, len(pitches) - 1):
                if InterpolationContour._is_turning_point_fantastic(pitches, i):
                    candidate_points_pitch.append(pitches[i])
                    candidate_points_time.append(onsets[i])
        else:
            # For longer melodies
            for i in range(2, len(pitches) - 2):
                if InterpolationContour._is_turning_point_fantastic(pitches, i):
                    candidate_points_pitch.append(pitches[i])
                    candidate_points_time.append(onsets[i])

        # Initialize turning points with first note
        turning_points_pitch = [pitches[0]]
        turning_points_time = [onsets[0]]

        # Find turning points
        if len(candidate_points_pitch) > 2:
            for i in range(1, len(pitches) - 1):
                if onsets[i] in candidate_points_time:
                    if pitches[i - 1] != pitches[i + 1]:
                        turning_points_pitch.append(pitches[i])
                        turning_points_time.append(onsets[i])

        # Add last note
        turning_points_pitch.append(pitches[-1])
        turning_points_time.append(onsets[-1])

        # Calculate gradients
        gradients = np.diff(turning_points_pitch) / np.diff(turning_points_time)

        # Calculate durations
        durations = np.diff(turning_points_time)

        # Create weighted gradients vector
        sample_rate = 10  # 10 samples per second
        samples_per_duration = abs(
            np.round(durations * sample_rate).astype(int)
        )
        interpolation_contour = np.repeat(gradients, samples_per_duration)

        return [float(x) for x in interpolation_contour]

    @staticmethod
    def _remove_repeated_notes(
        pitches: list[int], onsets: list[float]
    ) -> tuple[list[int], list[float]]:
        """Helper function to remove repeated notes, keeping only the middle occurrence.

        This is used for the AMADS method to produce the interpolated gradient values
        at the middle of a sequence of repeated notes, should there be a reversal
        between the repeated notes.
        """
        unique_pitches, unique_onsets = [], []
        i = 0
        while i < len(pitches):
            start_idx = i
            while i < len(pitches) - 1 and pitches[i + 1] == pitches[i]:
                i += 1
            mid_idx = start_idx + (i - start_idx) // 2
            unique_pitches.append(pitches[mid_idx])
            unique_onsets.append(onsets[mid_idx])
            i += 1
        return unique_pitches, unique_onsets

    @staticmethod
    def _calculate_amads_contour(
        pitches: list[int], onsets: list[float]
    ) -> list[float]:
        """
        Calculate the interpolation contour using the AMADS method.

        Utilises the helper function _remove_repeated_notes.
        """
        reversals_pitches = [pitches[0]]
        reversals_time = [onsets[0]]

        # Remove repeated notes
        pitches, onsets = InterpolationContour._remove_repeated_notes(
            pitches, onsets
        )

        # Find reversals
        for i in range(2, len(pitches)):
            if (
                pitches[i] < pitches[i - 1] > pitches[i - 2]
                or pitches[i] > pitches[i - 1] < pitches[i - 2]
            ):
                reversals_pitches.append(pitches[i - 1])
                reversals_time.append(onsets[i - 1])

        # Add last note
        reversals_pitches.append(pitches[-1])
        reversals_time.append(onsets[-1])

        # Calculate gradients
        gradients = np.diff(reversals_pitches) / np.diff(reversals_time)

        # Calculate durations
        durations = np.diff(reversals_time)

        # Create weighted gradients vector
        samples_per_duration = abs(np.round(durations * 10).astype(int))

        # Can't have a contour with less than 2 points
        if len(reversals_pitches) < 2:
            return [0.0]

        # If there are only 2 points, just use the gradient between them
        if len(reversals_pitches) == 2:
            gradient = reversals_pitches[1] - reversals_pitches[0]
            return [float(gradient / (reversals_time[1] - reversals_time[0]))]

        interpolation_contour = np.repeat(gradients, samples_per_duration)
        return [float(x) for x in interpolation_contour]

    @property
    def global_direction(self) -> int:
        """Calculate the global direction of the interpolation contour.

        Takes the sign of the sum of all contour values.
        Can be invoked for either FANTASTIC or AMADS method.

        Returns
        -------
        int
            1 if sum is positive, 0 if sum is zero, -1 if sum is negative

        Examples
        --------
        Flat overall contour direction (returns the same using FANTASTIC method)
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4])
        >>> ic.global_direction
        0

        Upwards contour direction (returns the same using FANTASTIC method)
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 65, 67], onsets=[0, 1, 2, 3, 4])
        >>> ic.global_direction
        1

        Downwards contour direction (returns the same using FANTASTIC method)
        >>> ic = InterpolationContour(pitches=[67, 65, 67, 62, 60], onsets=[0, 1, 2, 3, 4])
        >>> ic.global_direction
        -1
        """
        return int(np.sign(sum(self.contour)))

    @property
    def mean_gradient(self) -> float:
        """Calculate the absolute mean gradient of the interpolation contour.
        Can be invoked for either FANTASTIC or AMADS method.

        Returns
        -------
        float
            Mean of the absolute gradient values

        Examples
        --------
        Steps of 2 semitones per second
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4])
        >>> ic.mean_gradient
        2.0

        FANTASTIC method returns 0.0 for this example
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4], method="fantastic")
        >>> ic.mean_gradient
        0.0
        """
        return float(np.mean(np.abs(self.contour)))

    @property
    def gradient_std(self) -> float:
        """Calculate the standard deviation of the interpolation contour gradients.

        Can be invoked for either FANTASTIC or AMADS method.

        Returns
        -------
        float
            Standard deviation of the gradient values (by default, using Bessel's correction)

        Examples
        --------
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4])
        >>> round(ic.gradient_std, 7)
        2.0254787

        FANTASTIC method returns 0.0 for this example
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4], method="fantastic")
        >>> ic.gradient_std
        0.0
        """
        return float(np.std(self.contour, ddof=1))

    @property
    def direction_changes(self) -> float:
        """Calculate the proportion of interpolated gradient values that consistute
        a change in direction. For instance, a gradient value of
        -0.5 to 0.25 is a change in direction.
        Can be invoked for either FANTASTIC or AMADS method.

        Returns
        -------
        float
            Ratio of the number of changes in contour direction relative to the number
            of different interpolated gradient values

        Examples
        --------
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4])
        >>> ic.direction_changes
        1.0

        FANTASTIC method returns 0.0 for this example
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4], method="fantastic")
        >>> ic.direction_changes
        0.0
        """
        # Convert contour to numpy array for element-wise multiplication
        contour_array = np.array(self.contour)
        # Calculate products of consecutive gradients
        consecutive_products = contour_array[:-1] * contour_array[1:]

        # Get signs of products and count negative ones (direction changes)
        product_signs = np.sign(consecutive_products)
        direction_changes = np.sum(np.abs(product_signs[product_signs == -1]))

        # Count total gradient changes (where consecutive values are different)
        total_changes = np.sum(contour_array[:-1] != contour_array[1:])

        # Avoid division by zero
        if total_changes == 0:
            return 0.0

        return float(direction_changes / total_changes)

    @property
    def class_label(self) -> str:
        """Classify an interpolation contour into gradient categories.

        Can be invoked for either FANTASTIC or AMADS method.

        The contour is sampled at 4 equally spaced points and each gradient is
        normalized to units of pitch change per second
        (expressed in units of semitones per 0.25 seconds.)
        The result is then classified into one of 5 categories:

        - 'a': Strong downward (-2) - normalized gradient <= -1.45
        - 'b': Downward (-1) - normalized gradient between -1.45 and -0.45
        - 'c': Flat (0) - normalized gradient between -0.45 and 0.45
        - 'd': Upward (1) - normalized gradient between 0.45 and 1.45
        - 'e': Strong upward (2) - normalized gradient >= 1.45

        Returns
        -------
        str
            String of length 4 containing letters a-e representing the gradient
            categories at 4 equally spaced points in the contour

        Examples
        --------
        Upwards, then downwards contour
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4])
        >>> ic.class_label
        'ddbb'

        FANTASTIC method returns 'cccc' for this example, as though the contour is flat
        >>> ic = InterpolationContour(pitches=[60, 62, 64, 62, 60], onsets=[0, 1, 2, 3, 4], method="fantastic")
        >>> ic.class_label
        'cccc'
        """
        # Sample the contour at 4 equally spaced points
        # Get 4 equally spaced indices
        n = len(self.contour)
        indices = np.linspace(0, n - 1, 4, dtype=int)

        # Sample the contour at those indices
        sampled_points = [self.contour[i] for i in indices]

        # Normalize the gradients to a norm where value of 1 corresponds to a semitone
        # change in pitch over 0.25 seconds.
        # Given that base pitch and time units are 1 second and 1 semitone respectively,
        # just divide by 4
        norm_gradients = np.array(sampled_points) * 0.25
        classes = ""
        for grad in norm_gradients:
            if grad <= -1.45:
                classes += "a"  # strong down
            elif -1.45 < grad <= -0.45:
                classes += "b"  # down
            elif -0.45 < grad < 0.45:
                classes += "c"  # flat
            elif 0.45 <= grad < 1.45:
                classes += "d"  # up
            else:
                classes += "e"  # strong up

        return classes

    def plot(self, ax=None):
        """
        Plot the melody notes and the interpolation contour gradients.

        Displays two subplots (if ``ax`` is None):

        * **Top**: pitch values at their onset times as a scatter/step plot,
          with all original melody notes connected by lines.
        * **Bottom**: the interpolation contour — the piecewise-constant
          gradient values produced by `calculate_interpolation_contour`,
          plotted as a step function over normalised time (0–1).

        Parameters
        ----------
        ax : array-like of matplotlib.axes.Axes, optional
            A pair of Axes ``[ax_melody, ax_contour]`` to draw on.  If
            ``None``, a new figure with two vertically stacked subplots is
            created with ``figsize=(8, 5)``.

        Returns
        -------
        tuple[matplotlib.axes.Axes, matplotlib.axes.Axes]
            ``(ax_melody, ax_contour)`` — the two axes, suitable for further
            customisation or embedding in a larger figure.

        Raises
        ------
        ImportError
            If matplotlib is not installed.

        Examples
        --------
        >>> ic = InterpolationContour(
        ...     pitches=[60, 62, 64, 62, 60],
        ...     onsets=[0, 1, 2, 3, 4],
        ... )
        >>> ax_melody, ax_contour = ic.plot()
        """
        try:
            import matplotlib.pyplot as plt
        except ImportError as e:
            raise ImportError(
                "matplotlib is required for plotting. "
                "Install it with: pip install matplotlib"
            ) from e

        if ax is None:
            _, (ax_melody, ax_contour) = plt.subplots(
                2, 1, figsize=(8, 5), constrained_layout=True
            )
        else:
            ax_melody, ax_contour = ax

        # Top panel: all melody notes with connecting lines
        ax_melody.plot(
            self.onsets,
            self.pitches,
            color="steelblue",
            linewidth=1.2,
            zorder=2,
        )
        ax_melody.scatter(
            self.onsets,
            self.pitches,
            color="steelblue",
            s=50,
            linewidths=0.8,
            edgecolors="white",
            zorder=3,
            label="Notes",
        )
        ax_melody.set_xlabel("Onset time", fontsize=10)
        ax_melody.set_ylabel("MIDI pitch", fontsize=10)
        ax_melody.set_title("Melody", fontsize=11)
        ax_melody.grid(True, linestyle="--", alpha=0.4)
        ax_melody.spines[["top", "right"]].set_visible(False)

        # Bottom panel: interpolation contour
        contour = self.contour
        # The contour is sampled at a rate of 10 samples per unit time.
        # Build a normalised time axis over [0, 1] for display.
        n = len(contour)
        norm_time = np.linspace(0, 1, n, endpoint=False)

        ax_contour.step(
            norm_time,
            contour,
            where="post",
            color="tomato",
            linewidth=1.8,
            label=f"Interpolation contour ({self.method})",
        )
        ax_contour.axhline(0, color="black", linewidth=0.8, linestyle="--")
        ax_contour.set_xlabel("Normalised time", fontsize=10)
        ax_contour.set_ylabel("Gradient (semitones / s)", fontsize=10)
        ax_contour.set_title(
            f"Interpolation contour  "
            f"[direction={self.global_direction:+d}, "
            f"mean={self.mean_gradient:.2f}, "
            f"class={self.class_label}]",
            fontsize=11,
        )
        ax_contour.legend(fontsize=9)
        ax_contour.grid(True, linestyle="--", alpha=0.4)
        ax_contour.spines[["top", "right"]].set_visible(False)

        return ax_melody, ax_contour



"""compare two Score objects

"""

from math import isclose
from typing import cast

from amads.core.basics import (
    Chord,
    Clef,
    Event,
    EventGroup,
    KeySignature,
    Measure,
    Note,
    Part,
    Rest,
    Score,
    Staff,
    TimeSignature,
)

__author__ = "Roger B. Dannenberg"


def _time_map_error(msg, x1, x2, tm1, tm2):
    """print time_map error and data"""
    print(msg, x1, x2)
    print("    tm1: ", end="")
    tm1.show()
    print("    tm2: ", end="")
    tm2.show()


def compare_time_maps(tm1, tm2):
    """Compare two TimeMaps for equality."""
    if len(tm1.changes) != len(tm2.changes):
        return _time_map_error(
            "TimeMap lengths do not match:",
            len(tm1.changes),
            len(tm2.changes),
            tm1,
            tm2,
        )
    for mq1, mq2 in zip(tm1.changes, tm2.changes):
        if not isclose(mq1.time, mq2.time, abs_tol=1e-3):
            return _time_map_error(
                "TimeMap times do not match:", mq1.time, mq2.time, tm1, tm2
            )
        if not isclose(mq1.quarter, mq2.quarter, abs_tol=1e-3):
            return _time_map_error(
                "TimeMap quarters do not match:",
                mq1.quarter,
                mq2.quarter,
                tm1,
                tm2,
            )
    return True


def _score_compare_error(
    msg, score1, data1, score2, data2, data_phrase: str
) -> None:
    print(msg)
    print(f"  Score 1: {score1}", end="")
    if data1 == "skip":
        data1 = ""
    else:
        data1 = f" {data_phrase} {data1}"
    print(data1)
    print(f"  Score 2: {score2}", end="")
    if data2 == "skip":
        data2 = ""
    else:
        data2 = f" {data_phrase} {data2}"
    print(data2)


def _pitch_dim(event: Event) -> float:
    """compute a number for sorting events if onsets are equal

    We want Clef < KeySignature < Note, and Notes ordered by key_num
    and then alt.

    This function is not completely correct because if two key_nums
    are very close (on the order of 0.001), and have different alt
    values, they could be returned out of order, but in "normal" music,
    key_num values are separated by 1 or at least some audible difference.
    """
    if isinstance(event, Clef):
        return -10
    elif isinstance(event, KeySignature):
        return -5
    elif isinstance(event, Note):
        pitch = cast(Note, event).pitch
        return pitch.key_num - pitch.alt * 0.001
    return 0


def _all_but_rests_and_chords(measure: Measure, midi: bool) -> list[Event]:
    """
    returns content except rests and chords, but include chord content

    An additional hack is to remove KeySignature(onset=0, 0 flats) because
    this is implied but created anyway (only) by Music21.

    Similarly, if this is midi, Music21 may create Clefs, but Clef is not
    supported by MIDI so we remove them.

    The returned list is sorted in order of increasing time and pitch.
    """
    content = []
    for item in measure.content:
        if isinstance(item, Chord):
            for citem in item.content:  # move chord content to our content
                if not isinstance(citem, Rest):
                    content.append(citem)  # nested chords are NOT expanded
                    # but nested chords should not exist
        elif midi and isinstance(item, Clef):
            pass
        elif (
            isinstance(item, KeySignature)
            and item.onset == 0
            and item.key_sig == 0
        ):
            pass
        elif not isinstance(item, Rest):
            content.append(item)
    content.sort(key=lambda x: (x.onset, _pitch_dim(x)))
    return content


def scores_compare(score1: Score, score2: Score, midi: bool = False) -> bool:
    """Compare two Scores for equality.

    If they are different, print the first difference found.

    Ignores Rests, Chords and fields in non-standard subclasses of Event,
    but compares Score, Part, Staff, Measure, Note, TimeSignature,
    KeySignature, Clef.

    If midi, EventGroups are allowed to have non-matching durations because
    some MIDI readers fill out duration to even number of measures and others
    do not.  And Staff numbers do not have to match.

    Parameters
    ----------
    score1 : Event
        First score or event to compare
    score2 : Event
        Second score or event to compare

    Returns
    -------
    bool
        True if the scores/events are equal, False otherwise.

    <small>**Author**: Roger B. Dannenberg</small>
    """
    if not isinstance(score1, Event):
        print("Event or EventGroup from score1 is not an Event", score1)
        return False
    if not isinstance(score2, Event):
        print("Event or EventGroup from score2 is not an Event", score2)
        return False
    if not isclose(score1.onset, score2.onset, abs_tol=0.001):
        _score_compare_error(
            "Event onsets do not match:",
            score1,
            score1.onset,
            score2,
            score2.onset,
            "onset is",
        )
        return False
    dur_match = isclose(score1.duration, score2.duration, abs_tol=0.001)
    if isinstance(score1, EventGroup) and isinstance(score2, EventGroup):
        dur_match = dur_match or midi  # allow non-matching if midi
        content1 = score1.content
        content2 = score2.content
        if isinstance(score1, Measure) and isinstance(score2, Measure):
            # filter Rests and Chords from measures -- they're optional,
            # and MusicXML readers handle things differently
            content1 = _all_but_rests_and_chords(score1, midi)
            content2 = _all_but_rests_and_chords(score2, midi)

        if len(content1) != len(content2):
            _score_compare_error(
                "EventGroup content lengths do not match:",
                score1,
                len(content1),
                score2,
                len(content2),
                "length is",
            )
            print("    from score1:")
            for c in content1:
                print("        ", c)
            print("    from score2:", content2)
            for c in content2:
                print("        ", c)
            return False
        if isinstance(score1, Score) and isinstance(score2, Score):
            if score1._units_are_seconds != score2._units_are_seconds:
                _score_compare_error(
                    "Score units do not match:",
                    score1,
                    score1._units_are_seconds,
                    score2,
                    score2._units_are_seconds,
                    "_units_are_seconds is",
                )
                return False
            if compare_time_maps(score1.time_map, score2.time_map) is False:
                _score_compare_error(
                    "Score time maps do not match:",
                    score1,
                    score1.time_map,
                    score2,
                    score2.time_map,
                    "time_map is",
                )
        elif isinstance(score1, Part) and isinstance(score2, Part):
            # instruments should match, but in PrettyMIDI, if the
            # track name is empty, the instrument name is derived
            # from the General MIDI instrument corresponding to the
            # MIDI program used in the track, defaulting to 0,
            # which is "Acoustic Grand Piano":
            if (score1.instrument != score2.instrument) and (
                (not midi)
                or (
                    (score1.instrument is not None)
                    and (score2.instrument is not None)
                )
            ):
                _score_compare_error(
                    "Part MIDI instruments do not match:",
                    score1,
                    score1.instrument,
                    score2,
                    score2.instrument,
                    "instrument is",
                )
                return False
            if score1.number != score2.number:
                _score_compare_error(
                    "Part numbers do not match:",
                    score1,
                    score1.number,
                    score2,
                    score2.number,
                    "number is",
                )
                return False
        elif isinstance(score1, Staff) and isinstance(score2, Staff):
            # staff numbers do not need to match. Only Partitura numbers staffs
            # in MIDI files and numbering is 1, 2, ... within each Part, even
            # though we would expect
            if midi and score1.number != score2.number:
                _score_compare_error(
                    "Staff numbers do not match:",
                    score1,
                    score1.number,
                    score2,
                    score2.number,
                    "number is",
                )
                return False
        elif isinstance(score1, Measure) and isinstance(score2, Measure):
            if score1.number != score2.number:
                _score_compare_error(
                    "Measure numbers do not match:",
                    score1,
                    score1.number,
                    score2,
                    score2.number,
                    "number is",
                )
                return False
        for elem1, elem2 in zip(content1, content2):
            if not scores_compare(elem1, elem2, midi):
                return False
    else:  # both are Events
        # compare to makes sure they have the same class
        if score1.__class__ != score2.__class__:
            _score_compare_error(
                "Event classes do not match:",
                score1,
                score1.__class__,
                score2,
                score2.__class__,
                "class is",
            )
            return False
        if isinstance(score1, Note):
            score2 = cast(Note, score2)
            if (midi and (score1.key_num != score2.key_num)) or (
                (not midi) and (score1.pitch != score2.pitch)
            ):
                _score_compare_error(
                    "Note pitches do not match:",
                    score1,
                    score1.pitch,
                    score2,
                    score2.pitch,
                    "pitch is",
                )
                return False
            if score1.dynamic != score2.dynamic:
                _score_compare_error(
                    "Note velocities do not match:",
                    score1,
                    score1.dynamic,
                    score2,
                    score2.dynamic,
                    "dynamic is",
                )
                return False
            if score1.lyric != score2.lyric:
                _score_compare_error(
                    "Note lyrics do not match:",
                    score1,
                    score1.lyric,
                    score2,
                    score2.lyric,
                    "lyric is",
                )
                return False
            if score1.tie or score2.tie:
                if not score2.tie:
                    _score_compare_error(
                        "score1 has tie but score2 does not:",
                        score1,
                        score1.tie,
                        score2,
                        "skip",
                        "tie is",
                    )
                    return False
                elif not score1.tie:
                    _score_compare_error(
                        "score2 has tie but score1 does not:",
                        score1,
                        "skip",
                        score2,
                        score2.tie,
                        "tie is",
                    )
                    return False
                elif not scores_compare(score1.tie, score2.tie):
                    _score_compare_error(
                        "Tied notes do not match:",
                        score1,
                        score1.tie,
                        score2,
                        score2.tie,
                        "tied note is",
                    )
                    return False
                return True
        elif isinstance(score1, TimeSignature):
            score2 = cast(TimeSignature, score2)
            if score1.upper != score2.upper:
                _score_compare_error(
                    "TimeSignature uppers do not match:",
                    score1,
                    score1.upper,
                    score2,
                    score2.upper,
                    "upper is",
                )
                return False
            if score1.lower != score2.lower:
                _score_compare_error(
                    "TimeSignature lowers do not match:",
                    score1,
                    score1.lower,
                    score2,
                    score2.lower,
                    "lower is",
                )
                return False
        elif isinstance(score1, KeySignature):
            score2 = cast(KeySignature, score2)
            if score1.key_sig != score2.key_sig:
                _score_compare_error(
                    "KeySignature key_sigs do not match:",
                    score1,
                    score1.key_sig,
                    score2,
                    score2.key_sig,
                    "key_sig is",
                )
                return False
        elif isinstance(score1, Clef):
            score2 = cast(Clef, score2)
            if score1.clef != score2.clef:
                _score_compare_error(
                    "Clef types do not match:",
                    score1,
                    score1.clef,
                    score2,
                    score2.clef,
                    "clef is",
                )
                return False
        if not dur_match:
            _score_compare_error(
                "Event durations do not match:",
                score1,
                score1.duration,
                score2,
                score2.duration,
                "duration is",
            )
            return False

    return True



import sys

import pytest

coverage_args = ["--cov=./", "--cov-report=xml"]


def run_main_tests():
    """
    Run main tests with coverage.
    Tests are discovered following repo's pytest.ini file which says
    run tests in amads and tests named test_*, *_test, with classes
    Test* and function test_*.
    Assumes that the working directory is the root of the repository.
    """
    pytest_args = coverage_args
    sys.exit(pytest.main(pytest_args))



"""Test partimenti module.

This module tests the partimenti module functionality.

Created by Mark Gotham, 2019.

License: Creative Commons Attribution-ShareAlike 4.0 International License
https://creativecommons.org/licenses/by-sa/4.0/
"""

__author__ = "Mark Gotham"


# ------------------------------------------------------------------------------

from amads.schema.partimenti import all


def test_partimenti_length():
    """Test that partimento sections have consistent lengths ("stages").

    Each partimento should have melody, bass, and figures sections
    of equal length.
    """
    for p in all:
        assert len(p["melody"]) == len(p["bass"])
        assert len(p["melody"]) == len(p["figures"])



"""
This is a wrapper for key_cc to mimic the functionality of kkcc from
miditoolbox for convenience.

<small>**Author**: Tai Nakamura, Di Wang</small>

Reference
---------
https://citeseerx.ist.psu.edu/document?repid=rep1&type=pdf&doi=6e06906ca1ba0bf0ac8f2cb1a929f3be95eeadfa#page=68
"""

from itertools import chain
from typing import Tuple

from amads.core.basics import Score
from amads.pitch.key import profiles
from amads.pitch.key.key_cc import key_cc


def kkcc(
    score: Score,
    profile_name: str = "KRUMHANSL-KESSLER",
    salience_flag: bool = False,
) -> Tuple[float]:
    """
    kkcc wrapper on key_cc that provides the exact behavior of miditoolbox kkcc

    This module:

      1. Provides 3 string options for profile names
      2. maps the `profile_name` option to the relevant profile and attribute
         name list combination for key_cc, replicating the behavior of
         the relevant kkcc function call in miditoolbox.

    Parameters
    ----------
    score : Score
        The musical score to analyze.
    profile_name : str
        String argument denoting the relevant miditoolbox
        string option for kkcc. Must be one of "KRUMHANSL-KESSLER",
        "TEMPERLEY", or "ALBRECHT-SHANAHAN".
    salience_flag : bool
        If True, apply salience weighting to the pitch-class
        according to Huron & Parncutt (1993).

    Returns
    -------
    tuple[float, ...]
        This denotes the 12 major correlation coefficients and 12 minor correlation
        coefficients from C to B in both major and minor keys, respectively (dim=24).

    Raises
    ------
    ValueError
        If the score is not a valid Score object or if the profile_name is invalid.

    See Also
    --------
    key_cc
    """
    if not isinstance(score, Score):
        raise ValueError("invalid score type!")
    # default is krumhansl kessler, and is what profile_name is set to by default
    profile = None
    attribute_list = None
    if profile_name == "KRUMHANSL-KESSLER":
        profile = profiles.krumhansl_kessler
        attribute_list = ["major", "minor"]
    elif profile_name == "TEMPERLEY":
        profile = profiles.temperley
        attribute_list = ["major", "minor"]
    elif profile_name == "ALBRECHT-SHANAHAN":
        profile = profiles.albrecht_shanahan
        attribute_list = ["major", "minor"]
    else:
        raise ValueError(f'profile_name = "{profile_name}" is not valid')

    # these checks are paranoia mainly to prevent future changes
    # from breaking the code after
    assert not (profile is None or attribute_list is None)
    assert isinstance(profile, profiles.KeyProfile)
    assert len(attribute_list) == 2

    corrcoef_pairs = key_cc(score, profile, attribute_list, salience_flag)
    # check integrity of corrcoef correspondences and whether or not they abide
    # to the output agreed on in key_cc
    assert len(corrcoef_pairs) == len(attribute_list)
    assert all(
        attr_name == target_attr and len(coefs) == 12
        for ((attr_name, coefs), target_attr) in zip(
            corrcoef_pairs, attribute_list
        )
    )

    # pattern match, then collect individual coefficients into
    # a single tuple to get our final corrcoefs
    nested_coefs_iter = (coefs for (_, coefs) in corrcoef_pairs)
    corrcoefs = tuple(chain.from_iterable(nested_coefs_iter))

    return corrcoefs
